@@ -18,7 +18,7 @@
  * Boston, MA 02110-1301 USA.
  *
  * Copyright (C) 2007 - 2008 Novell, Inc.
- * Copyright (C) 2007 - 2008 Red Hat, Inc.
+ * Copyright (C) 2007 - 2010 Red Hat, Inc.
  */
 
 #include <string.h>
@@ -49,16 +49,20 @@ typedef struct {
 	DBusGProxy *proxy;
 
 	char *iface;
+	char *ip_iface;
 	char *udi;
 	char *driver;
 	guint32 capabilities;
 	gboolean managed;
+	gboolean firmware_missing;
 	NMIP4Config *ip4_config;
 	gboolean null_ip4_config;
 	NMDHCP4Config *dhcp4_config;
 	gboolean null_dhcp4_config;
 	NMIP6Config *ip6_config;
 	gboolean null_ip6_config;
+	NMDHCP6Config *dhcp6_config;
+	gboolean null_dhcp6_config;
 	NMDeviceState state;
 
 	GUdevClient *client;
@@ -73,12 +77,15 @@ enum {
 	PROP_DRIVER,
 	PROP_CAPABILITIES,
 	PROP_MANAGED,
+	PROP_FIRMWARE_MISSING,
 	PROP_IP4_CONFIG,
 	PROP_DHCP4_CONFIG,
 	PROP_IP6_CONFIG,
 	PROP_STATE,
 	PROP_PRODUCT,
 	PROP_VENDOR,
+	PROP_DHCP6_CONFIG,
+	PROP_IP_INTERFACE,
 
 	LAST_PROP
 };
@@ -220,19 +227,62 @@ demarshal_ip6_config (NMObject *object, GParamSpec *pspec, GValue *value, gpoint
 	return TRUE;
 }
 
+static gboolean
+demarshal_dhcp6_config (NMObject *object, GParamSpec *pspec, GValue *value, gpointer field)
+{
+	NMDevicePrivate *priv = NM_DEVICE_GET_PRIVATE (object);
+	const char *path;
+	NMDHCP6Config *config = NULL;
+	DBusGConnection *connection;
+
+	if (!G_VALUE_HOLDS (value, DBUS_TYPE_G_OBJECT_PATH))
+		return FALSE;
+
+	priv->null_dhcp6_config = FALSE;
+
+	path = g_value_get_boxed (value);
+	if (path) {
+		if (!strcmp (path, "/"))
+			priv->null_dhcp6_config = TRUE;
+		else {
+			config = NM_DHCP6_CONFIG (_nm_object_cache_get (path));
+			if (config)
+				config = g_object_ref (config);
+			else {
+				connection = nm_object_get_connection (object);
+				config = NM_DHCP6_CONFIG (nm_dhcp6_config_new (connection, path));
+			}
+		}
+	}
+
+	if (priv->dhcp6_config) {
+		g_object_unref (priv->dhcp6_config);
+		priv->dhcp6_config = NULL;
+	}
+
+	if (config)
+		priv->dhcp6_config = config;
+
+	_nm_object_queue_notify (object, NM_DEVICE_DHCP6_CONFIG);
+	return TRUE;
+}
+
 static void
 register_for_property_changed (NMDevice *device)
 {
 	NMDevicePrivate *priv = NM_DEVICE_GET_PRIVATE (device);
 	const NMPropertiesChangedInfo property_changed_info[] = {
-		{ NM_DEVICE_UDI,          _nm_object_demarshal_generic, &priv->udi },
-		{ NM_DEVICE_INTERFACE,    _nm_object_demarshal_generic, &priv->iface },
-		{ NM_DEVICE_DRIVER,       _nm_object_demarshal_generic, &priv->driver },
-		{ NM_DEVICE_CAPABILITIES, _nm_object_demarshal_generic, &priv->capabilities },
-		{ NM_DEVICE_MANAGED,      _nm_object_demarshal_generic, &priv->managed },
-		{ NM_DEVICE_IP4_CONFIG,   demarshal_ip4_config,        &priv->ip4_config },
-		{ NM_DEVICE_DHCP4_CONFIG, demarshal_dhcp4_config,      &priv->dhcp4_config },
-		{ NM_DEVICE_IP6_CONFIG,   demarshal_ip6_config,        &priv->ip6_config },
+		{ NM_DEVICE_UDI,              _nm_object_demarshal_generic, &priv->udi },
+		{ NM_DEVICE_INTERFACE,        _nm_object_demarshal_generic, &priv->iface },
+		{ NM_DEVICE_IP_INTERFACE,     _nm_object_demarshal_generic, &priv->ip_iface },
+		{ NM_DEVICE_DRIVER,           _nm_object_demarshal_generic, &priv->driver },
+		{ NM_DEVICE_CAPABILITIES,     _nm_object_demarshal_generic, &priv->capabilities },
+		{ NM_DEVICE_MANAGED,          _nm_object_demarshal_generic, &priv->managed },
+		{ NM_DEVICE_FIRMWARE_MISSING, _nm_object_demarshal_generic, &priv->firmware_missing },
+		{ NM_DEVICE_IP4_CONFIG,       demarshal_ip4_config,         &priv->ip4_config },
+		{ NM_DEVICE_DHCP4_CONFIG,     demarshal_dhcp4_config,       &priv->dhcp4_config },
+		{ NM_DEVICE_IP6_CONFIG,       demarshal_ip6_config,         &priv->ip6_config },
+		{ NM_DEVICE_DHCP6_CONFIG,     demarshal_dhcp6_config,       &priv->dhcp6_config },
 		{ NULL },
 	};
 
@@ -318,6 +368,8 @@ dispose (GObject *object)
 		g_object_unref (priv->dhcp4_config);
 	if (priv->ip6_config)
 		g_object_unref (priv->ip6_config);
+	if (priv->dhcp6_config)
+		g_object_unref (priv->dhcp6_config);
 	if (priv->client)
 		g_object_unref (priv->client);
 
@@ -330,6 +382,7 @@ finalize (GObject *object)
 	NMDevicePrivate *priv = NM_DEVICE_GET_PRIVATE (object);
 
 	g_free (priv->iface);
+	g_free (priv->ip_iface);
 	g_free (priv->udi);
 	g_free (priv->driver);
 	g_free (priv->product);
@@ -353,6 +406,9 @@ get_property (GObject *object,
 	case PROP_INTERFACE:
 		g_value_set_string (value, nm_device_get_iface (device));
 		break;
+	case PROP_IP_INTERFACE:
+		g_value_set_string (value, nm_device_get_ip_iface (device));
+		break;
 	case PROP_DRIVER:
 		g_value_set_string (value, nm_device_get_driver (device));
 		break;
@@ -362,6 +418,9 @@ get_property (GObject *object,
 	case PROP_MANAGED:
 		g_value_set_boolean (value, nm_device_get_managed (device));
 		break;
+	case PROP_FIRMWARE_MISSING:
+		g_value_set_boolean (value, nm_device_get_firmware_missing (device));
+		break;
 	case PROP_IP4_CONFIG:
 		g_value_set_object (value, nm_device_get_ip4_config (device));
 		break;
@@ -371,6 +430,9 @@ get_property (GObject *object,
 	case PROP_IP6_CONFIG:
 		g_value_set_object (value, nm_device_get_ip6_config (device));
 		break;
+	case PROP_DHCP6_CONFIG:
+		g_value_set_object (value, nm_device_get_dhcp6_config (device));
+		break;
 	case PROP_STATE:
 		g_value_set_uint (value, nm_device_get_state (device));
 		break;
@@ -466,6 +528,20 @@ nm_device_class_init (NMDeviceClass *device_class)
 						  FALSE,
 						  G_PARAM_READABLE));
 
+	/**
+	 * NMDevice:firmware-missing:
+	 *
+	 * When %TRUE indicates the device is likely missing firmware required
+	 * for its operation.
+	 **/
+	g_object_class_install_property
+		(object_class, PROP_FIRMWARE_MISSING,
+		 g_param_spec_boolean (NM_DEVICE_FIRMWARE_MISSING,
+						  "FirmwareMissing",
+						  "Firmware missing",
+						  FALSE,
+						  G_PARAM_READABLE));
+
 	/**
 	 * NMDevice:ip4-config:
 	 *
@@ -505,6 +581,19 @@ nm_device_class_init (NMDeviceClass *device_class)
 		                      NM_TYPE_IP6_CONFIG,
 		                      G_PARAM_READABLE));
 
+	/**
+	 * NMDevice:dhcp6-config:
+	 *
+	 * The #NMDHCP6Config of the device.
+	 **/
+	g_object_class_install_property
+		(object_class, PROP_DHCP6_CONFIG,
+		 g_param_spec_object (NM_DEVICE_DHCP6_CONFIG,
+						  "DHCP6 Config",
+						  "DHCP6 Config",
+						  NM_TYPE_DHCP6_CONFIG,
+						  G_PARAM_READABLE));
+
 	/**
 	 * NMDevice:state:
 	 *
@@ -664,6 +753,33 @@ nm_device_get_iface (NMDevice *device)
 	return priv->iface;
 }
 
+/**
+ * nm_device_get_ip_iface:
+ * @device: a #NMDevice
+ *
+ * Gets the IP interface name of the #NMDevice over which IP traffic flows
+ * when the device is in the ACTIVATED state.
+ *
+ * Returns: the IP traffic interface of the device. This is the internal string
+ * used by the device, and must not be modified.
+ **/
+const char *
+nm_device_get_ip_iface (NMDevice *device)
+{
+	NMDevicePrivate *priv;
+
+	g_return_val_if_fail (NM_IS_DEVICE (device), NULL);
+
+	priv = NM_DEVICE_GET_PRIVATE (device);
+	if (!priv->ip_iface) {
+		priv->ip_iface = _nm_object_get_string_property (NM_OBJECT (device),
+		                                                 NM_DBUS_INTERFACE_DEVICE,
+		                                                 "IpInterface");
+	}
+
+	return priv->ip_iface;
+}
+
 /**
  * nm_device_get_udi:
  * @device: a #NMDevice
@@ -767,6 +883,33 @@ nm_device_get_managed (NMDevice *device)
 	return priv->managed;
 }
 
+/**
+ * nm_device_get_firmware_missing:
+ * @device: a #NMDevice
+ *
+ * Indicates that firmware required for the device's operation is likely
+ * to be missing.
+ *
+ * Returns: %TRUE if firmware required for the device's operation is likely
+ * to be missing.
+ **/
+gboolean
+nm_device_get_firmware_missing (NMDevice *device)
+{
+	NMDevicePrivate *priv;
+
+	g_return_val_if_fail (NM_IS_DEVICE (device), 0);
+
+	priv = NM_DEVICE_GET_PRIVATE (device);
+	if (!priv->firmware_missing) {
+		priv->firmware_missing = _nm_object_get_boolean_property (NM_OBJECT (device),
+		                                                          NM_DBUS_INTERFACE_DEVICE,
+		                                                          "FirmwareMissing");
+	}
+
+	return priv->firmware_missing;
+}
+
 /**
  * nm_device_get_ip4_config:
  * @device: a #NMDevice
@@ -870,6 +1013,41 @@ nm_device_get_ip6_config (NMDevice *device)
 	return priv->ip6_config;
 }
 
+/**
+ * nm_device_get_dhcp6_config:
+ * @device: a #NMDevice
+ *
+ * Gets the current #NMDHCP6Config associated with the #NMDevice.
+ *
+ * Returns: the #NMDHCPConfig or %NULL if the device is not activated or not
+ * using DHCP.
+ **/
+NMDHCP6Config *
+nm_device_get_dhcp6_config (NMDevice *device)
+{
+	NMDevicePrivate *priv;
+	char *path;
+	GValue value = { 0, };
+
+	g_return_val_if_fail (NM_IS_DEVICE (device), NULL);
+
+	priv = NM_DEVICE_GET_PRIVATE (device);
+	if (priv->dhcp6_config)
+		return priv->dhcp6_config;
+	if (priv->null_dhcp6_config)
+		return NULL;
+
+	path = _nm_object_get_object_path_property (NM_OBJECT (device), NM_DBUS_INTERFACE_DEVICE, "Dhcp6Config");
+	if (path) {
+		g_value_init (&value, DBUS_TYPE_G_OBJECT_PATH);
+		g_value_take_boxed (&value, path);
+		demarshal_dhcp6_config (NM_OBJECT (device), NULL, &value, &priv->dhcp6_config);
+		g_value_unset (&value);
+	}
+
+	return priv->dhcp6_config;
+}
+
 /**
  * nm_device_get_state:
  * @device: a #NMDevice
@@ -954,7 +1132,7 @@ nm_device_update_description (NMDevice *device)
 {
 	NMDevicePrivate *priv;
 	const char *subsys[3] = { "net", "tty", NULL };
-	GUdevDevice *udev_device = NULL, *tmpdev;
+	GUdevDevice *udev_device = NULL, *tmpdev, *olddev;
 	const char *ifname;
 	guint32 count = 0;
 	const char *vendor, *model;
@@ -972,9 +1150,8 @@ nm_device_update_description (NMDevice *device)
 	if (!ifname)
 		return;
 
-	if (NM_IS_DEVICE_ETHERNET (device) || NM_IS_DEVICE_WIFI (device))
-		udev_device = g_udev_client_query_by_subsystem_and_name (priv->client, "net", ifname);
-	else if (NM_IS_GSM_DEVICE (device) || NM_IS_CDMA_DEVICE (device))
+	udev_device = g_udev_client_query_by_subsystem_and_name (priv->client, "net", ifname);
+	if (!udev_device)
 		udev_device = g_udev_client_query_by_subsystem_and_name (priv->client, "tty", ifname);
 	if (!udev_device)
 		return;
@@ -987,7 +1164,11 @@ nm_device_update_description (NMDevice *device)
 	/* Walk up the chain of the device and its parents a few steps to grab
 	 * vendor and device ID information off it.
 	 */
-	tmpdev = udev_device;
+
+	/* Ref the device again because we have to unref it each iteration,
+	 * as g_udev_device_get_parent() returns a ref-ed object.
+	 */
+	tmpdev = g_object_ref (udev_device);
 	while ((count++ < 3) && tmpdev && (!priv->vendor || !priv->product)) {
 		if (!priv->vendor)
 			priv->vendor = get_decoded_property (tmpdev, "ID_VENDOR_ENC");
@@ -995,11 +1176,23 @@ nm_device_update_description (NMDevice *device)
 		if (!priv->product)
 			priv->product = get_decoded_property (tmpdev, "ID_MODEL_ENC");
 
+		olddev = tmpdev;
 		tmpdev = g_udev_device_get_parent (tmpdev);
+		g_object_unref (olddev);
 	}
 
+	/* Unref the last device if we found what we needed before running out
+	 * of parents.
+	 */
+	if (tmpdev)
+		g_object_unref (tmpdev);
+
 	/* If we didn't get strings directly from the device, try database strings */
-	tmpdev = udev_device;
+
+	/* Again, ref the original device as we need to unref it every iteration
+	 * since g_udev_device_get_parent() returns a refed object.
+	 */
+	tmpdev = g_object_ref (udev_device);
 	count = 0;
 	while ((count++ < 3) && tmpdev && (!priv->vendor || !priv->product)) {
 		if (!priv->vendor) {
@@ -1014,9 +1207,20 @@ nm_device_update_description (NMDevice *device)
 				priv->product = g_strdup (model);
 		}
 
+		olddev = tmpdev;
 		tmpdev = g_udev_device_get_parent (tmpdev);
+		g_object_unref (olddev);
 	}
 
+	/* Unref the last device if we found what we needed before running out
+	 * of parents.
+	 */
+	if (tmpdev)
+		g_object_unref (tmpdev);
+
+	/* Balance the initial g_udev_client_query_by_subsystem_and_name() */
+	g_object_unref (udev_device);
+
 	_nm_object_queue_notify (NM_OBJECT (device), NM_DEVICE_VENDOR);
 	_nm_object_queue_notify (NM_OBJECT (device), NM_DEVICE_PRODUCT);
 }
